/*
 *  linux/fs/bfs/file.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 *
 *  from
 *
 *  linux/fs/minix/file.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>


ssize_t bfs_file_read (struct file *, char *,  size_t, loff_t *);
static bfs_inode_addr bfs_startpos_from_ds (struct super_block *,
	bfs_data_stream *, off_t);
static int bfs_read_indirect_block (struct super_block *, const bfs_inode_addr,
	const int, bfs_inode_addr *);
static int bfs_read_double_indirect_block (struct super_block *,
	const bfs_inode_addr, const int, bfs_inode_addr *);
static ssize_t bfs_file_write (struct file *, const char *, size_t, loff_t *);


static struct file_operations bfs_file_ops =
{
	NULL,				/* lseek */
	bfs_file_read,			/* read */
#ifdef CONFIG_BFS_FS_RW
	bfs_file_write,			/* write */
#else
	NULL,				/* write */
#endif
	NULL,				/* readdir - bad */
	NULL,				/* poll - default */
	NULL,				/* ioctl - default */
	generic_file_mmap,		/* mmap */
	NULL,
	NULL,				/* flush */
	NULL,				/* release */
	file_fsync,			/* fsync */
};

struct inode_operations bfs_file_inode_operations =
{
	&bfs_file_ops,			/* default file operations */
	NULL,				/* create */
	NULL,				/* lookup */
	NULL,				/* link */
	NULL,				/* unlink */
	NULL,				/* symlink */
	NULL,				/* mkdir */
	NULL,				/* rmdir */
	NULL,				/* mknod */
	NULL,				/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	generic_readpage,		/* readpage */
	NULL,				/* writepage */
	&bfs_bmap,			/* bmap */
	NULL,				/* truncate */
	NULL,				/* permission */
	NULL				/* smap */
};


ssize_t bfs_file_read(struct file *filp, char *buf,  size_t count, loff_t *ppos)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb = inode->i_sb;
	bfs_data_stream *    ds = &inode->u.bfs_i.i_data.ds;
	bfs_inode_addr       iaddr;
	bfs_inode_addr       iaddr2;
	off_t                pos = *ppos;
	off_t                sum = 0;
	int                  i;
	int                  j;
	off_t                read_count;
	off_t                offset;

	BFS_OUTPUT (("---> bfs_file_read() "
		"inode %lu count %lu ppos %Lu\n",
		inode->i_ino, (__u32) count, *ppos));

	/*
	 * Get start position
	 */

	for (i = 0; ;) {
		iaddr = bfs_read_data_stream (sb, ds, &i);
		if (BFS_IS_EMPTY_IADDR(&iaddr)) {
			BFS_OUTPUT ((" end data_stream\n"));
			break;
		}

		if ((pos >= sum)
			&& (pos < sum + iaddr.len * sb->u.bfs_sb.block_size)) {

			iaddr.start += (pos - sum) / sb->u.bfs_sb.block_size;
			iaddr.len -= (pos - sum) / sb->u.bfs_sb.block_size;
			break;
		}
		sum += iaddr.len * sb->u.bfs_sb.block_size;
	}

	read_count = 0;
	offset = (pos - sum) % sb->u.bfs_sb.block_size;
	iaddr2 = iaddr;

	count = inode->i_size > pos + count ? count : inode->i_size - pos;

	for ( j = 0; j < iaddr.len; j++) {

		struct buffer_head * bh;

		bh = bfs_bread2 (sb, iaddr2);
		if (!bh)
			return -EBADF;

		BFS_OUTPUT ((" read_count %Ld offset %Ld\n",
			read_count, offset));

		if (count > sb->u.bfs_sb.block_size) {

			memcpy (buf + read_count, bh->b_data + offset,
				sb->u.bfs_sb.block_size - offset);

			*ppos += sb->u.bfs_sb.block_size - offset;
			count -= sb->u.bfs_sb.block_size - offset;
			read_count += sb->u.bfs_sb.block_size - offset;
			offset = 0;

			iaddr2.start += 1;

			brelse (bh);

		} else {

			memcpy (buf + read_count, bh->b_data + offset, count);

			*ppos += count;
			read_count += count;
	
			brelse (bh);
			break;
		}
	}

	ppos += read_count;

	BFS_OUTPUT (("<--- bfs_file_read() "
		"return value %d, ppos %Ld\n", read_count, *ppos));

	return read_count;
}

/*
 * Read indirect block
 */

static int bfs_read_indirect_block (struct super_block * sb,
	const bfs_inode_addr indirect, const int pos, 
	bfs_inode_addr * iaddr)
{
	bfs_inode_addr *     array;
	struct buffer_head * bh;
	bfs_inode_addr       addr = indirect;
	int                  p = pos;

	BFS_OUTPUT (("---> bfs_read_indirect_block()\n"));

	/*
	 * explore nessealy block
	 */

	addr.start += p / BFS_BLOCK_PER_INODE(sb);
	addr.len -= p / BFS_BLOCK_PER_INODE(sb);

	if (addr.len < 1) {
		goto error;
	}

	p = p % BFS_BLOCK_PER_INODE(sb);

	bh = bfs_bread2 (sb, addr);
	if (!bh) {
		goto error;
	}

	array = (bfs_inode_addr *) bh->b_data;

	/*
	 * Is this block inode address??
	 */

	if (!BFS_IS_EMPTY_IADDR(&array[p % BFS_BLOCK_PER_INODE(sb)])) {
		*iaddr = array[p % BFS_BLOCK_PER_INODE(sb)];
	} else {
		brelse (bh);
		goto error;
	}

	brelse (bh);

	BFS_OUTPUT (("<--- bfs_read_indirect_block()\n"));

	return 0;

error:
	printk ("BFS: cannot read indirect block\n");
	return -EBADF;
}

/*
 * bfs_read_double_indirect_block
 *
 * description:
 *  Read double-indirect block
 *
 * parameter:
 *  sb              ... super block
 *  double_indirect ... inode address of double-indirect block
 *  pos             ... position of double-indirect block
 *                       (*pos - max_direct_block - max_indirect_block)
 *
 * return value:
 *  0 ... sucess
 */

static int bfs_read_double_indirect_block (struct super_block * sb,
        const bfs_inode_addr double_indirect, const int pos,
	bfs_inode_addr * iaddr)
{
        struct buffer_head *  bh_indirect;
        int                   p = pos;
        bfs_inode_addr *      indirects;
	bfs_inode_addr        addr = double_indirect;
	int                   i;

	BFS_OUTPUT (("---> bfs_read_double_indirect_block() \n"));

        while (addr.len) {
                bh_indirect = bfs_bread2 (sb, addr);

                if (!bh_indirect) {
                        BFS_OUTPUT (("cannot read double-indirect block "
                                "[%lu, %u, %u]\n",
                                double_indirect.allocation_group,
                                double_indirect.start, double_indirect.len));
                        return -EBADF;
                }

                addr.start++;
                addr.len--;

                indirects = (bfs_inode_addr *) bh_indirect->b_data;

                for (i = 0; i < BFS_BLOCK_PER_INODE(sb); i++) {

                        if (p < indirects[i].len * BFS_BLOCK_PER_INODE(sb)) {

                                /*
                                 * find block!
                                 */

				int err;

				err = bfs_read_indirect_block (sb,
					indirects[i], p, iaddr);
				brelse (bh_indirect);

				return err;
                        }

                        p -= indirects[i].len * BFS_BLOCK_PER_INODE(sb);
                }

                brelse(bh_indirect);
        }

        /*
         * cannot find block...
         */

        BFS_OUTPUT (("cannot find block...\n"));

        return -EIO;
}


bfs_inode_addr bfs_read_data_stream (struct super_block * sb,
	bfs_data_stream * ds, int * pos)
{
        bfs_inode_addr       iaddr = {0, 0, 0};

	BFS_OUTPUT (("---> bfs_read_data_stream() "
		"pos %d\n", *pos));

        if( *pos < 0 )
                return iaddr;

        if (*pos < ds->max_direct_range) {

		/*
		 * This position is in direct block.
		 */

		BFS_OUTPUT ((" read in direct block [%lu, %u, %u]\n",
			ds->direct[*pos].allocation_group,
			ds->direct[*pos].start, ds->direct[*pos].len));

		/*
		 * Is this block inode address??
		 */

                if (!BFS_IS_EMPTY_IADDR(&ds->direct[*pos])) {
			iaddr = ds->direct[(*pos)++];
		}
        } else if (*pos < ds->max_indirect_range + ds->max_direct_range) {

		/*
		 * This position is in in-direct block.
		 */

		int  p = *pos - ds->max_direct_range;
		
		BFS_OUTPUT ((" read in indirect block [%lu, %u, %u]\n",
			ds->indirect.allocation_group, ds->indirect.start,
			ds->indirect.len));

		if (!bfs_read_indirect_block (sb, ds->indirect, p, &iaddr)) {
			(*pos)++;
		} 
        } else if (*pos < ds->max_double_indirect_range
		+ ds->max_direct_range + ds->max_indirect_range) {

		/*
		 * This position is in double-in-direct block.
		 */

		int  p = *pos - ds->max_direct_range - ds->max_indirect_range;

		BFS_OUTPUT ((" read in double indirect block\n"));

		if (!bfs_read_double_indirect_block (sb, ds->double_indirect,
			p, &iaddr)) {

			(*pos)++;
		} 
        } else {
		/*
		 * out of range block
		 */
	}

	if (BFS_IS_EMPTY_IADDR(&iaddr))
		*pos = 0;

        return iaddr;
}

/*
 * Get inode address of start position form data stream
 */

static bfs_inode_addr bfs_startpos_from_ds (struct super_block * sb,
					bfs_data_stream * ds, off_t pos)
{
	bfs_inode_addr iaddr;
	off_t          sum = 0;
	int            i = 0;

	BFS_OUTPUT (("---> bfs_startpos_from_ds()\n"));

	/*
	 * Get start position
	 */

	while(1) {
		iaddr = bfs_read_data_stream (sb, ds, &i);
		if (BFS_IS_EMPTY_IADDR(&iaddr)) {
			BFS_OUTPUT ((" end data_stream\n"));
			break;
		}

		if ((pos >= sum)
			&& (pos < sum + iaddr.len * sb->u.bfs_sb.block_size)) {
			iaddr.start += (pos - sum) / sb->u.bfs_sb.block_size;
			iaddr.len -= (pos - sum) / sb->u.bfs_sb.block_size;
			break;
		}
		sum += iaddr.len * sb->u.bfs_sb.block_size;
	}

	BFS_OUTPUT (("<--- bfs_startpos_from_ds()\n"));

	return iaddr;
}

#ifdef CONFIG_BFS_RW 
static ssize_t bfs_file_write (struct file * filp, const char * buf,
                                size_t count, loff_t *ppos)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb;
//	struct buffer_head * bh;
	bfs_data_stream *    ds;
	bfs_inode_addr       iaddr;
	loff_t               pos;
	ssize_t              written;

	if (!inode) {
		return -EINVAL;
	}

	if (!S_ISREG(inode->i_mode)) {
		return -EINVAL;
	}

	sb = inode->i_sb;
	ds = &inode->u.bfs_i.i_data.ds;

	if (filp->f_flags & O_APPEND)
		pos = inode->i_size;
	else
		pos = *ppos;

	iaddr = bfs_startpos_from_ds (sb, ds, pos);
	written = 0;

	if (pos > inode->i_size)
		inode->i_size = pos;

	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	*ppos = pos;
	mark_inode_dirty(inode);

	return written;
}
#endif
