/*
 *  linux/bfs/fs/index.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 *
 *
 *  Index functions for BFS.
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/slab.h>


static void bfs_convert_index_entry (int, bfs_index_entry *, bfs_index_entry *);
static void bfs_convert_index_node (int, bfs_index_node *, bfs_index_node *);

/*
 * Convert little-edian and big-edian.
 */

static void bfs_convert_index_entry (int fstype, bfs_index_entry * entry,
	bfs_index_entry * out)
{
	switch (fstype)
	{
	case BFS_X86:
		out->magic = le32_to_cpu(entry->magic);
		out->node_size = le32_to_cpu(entry->node_size);
		out->max_number_of_levels =
			le32_to_cpu(entry->max_number_of_levels);
		out->data_type = le32_to_cpu(entry->data_type);
		out->root_node_pointer =
			le64_to_cpu(entry->root_node_pointer);
		out->free_node_pointer =
			le64_to_cpu(entry->free_node_pointer);
		out->maximum_size = le64_to_cpu(entry->maximum_size);
		break;

	case BFS_PPC:
		out->magic = be32_to_cpu(entry->magic);
		out->node_size = be32_to_cpu(entry->node_size);
		out->max_number_of_levels =
			be32_to_cpu(entry->max_number_of_levels);
		out->data_type = be32_to_cpu(entry->data_type);
		out->root_node_pointer =
			be64_to_cpu(entry->root_node_pointer);
		out->free_node_pointer =
			be64_to_cpu(entry->free_node_pointer);
		out->maximum_size = be64_to_cpu(entry->maximum_size);
		break;
	}
}


static void bfs_convert_index_node (int fstype, bfs_index_node * node,
	bfs_index_node * out)
{
	switch (fstype)
	{
	case BFS_X86:
		out->left = le64_to_cpu (node->left);
		out->right = le64_to_cpu (node->right);
		out->overflow = le64_to_cpu (node->overflow);
		out->all_key_count = le16_to_cpu (node->all_key_count);
		out->all_key_length = le16_to_cpu (node->all_key_length);
		break;

	case BFS_PPC:
		out->left = be64_to_cpu (node->left);
		out->right = be64_to_cpu (node->right);
		out->overflow = be64_to_cpu (node->overflow);
		out->all_key_count = be16_to_cpu (node->all_key_count);
		out->all_key_length = be16_to_cpu (node->all_key_length);
		break;
	}
}


/*
 * bfs_read_index_node
 *
 * description:
 *  read index node from inode.
 *
 * parameter:
 *  inode  ... address of index node
 *  sb     ... super block
 *  flags  ... 0 ... inode is index entry
 *             1 ... inode is index node
 *  offset ... offset into index address
 */

struct buffer_head * bfs_read_index_node (bfs_inode_addr inode,
	struct super_block * sb, int flags, bfs_off_t * offset)
{
	struct buffer_head * bh;
	bfs_inode_addr       iaddr;
	bfs_index_entry *    entry;

	BFS_OUTPUT (("---> bfs_read_index_node() "
		"index [%lu, %u, %u]\n",
		inode.allocation_group, inode.start, inode.len));

	if (!flags) {

		/*
		 * read index entry
		 */

		bh = bfs_bread2 (sb, inode);
		if (!bh) {
			printk (KERN_ERR "BFS: cannot read index entry.\n");
			return NULL;
		}
#ifndef CONFIG_BFS_CONV
		entry = (bfs_index_entry *) bh->b_data;
#else
		entry = (bfs_index_entry *) __getname();;
		if (!entry) {
			printk (KERN_ERR "BFS: cannot allocate memory\n");
			return NULL;
		}

		/*
		 * Convert little-edian and big-edian.
		 */

		bfs_convert_index_entry (sb->u.bfs_sb.mount_opts.bfs_type,
			(bfs_index_entry *) bh->b_data, entry);
#endif
		BFS_DUMP_INDEX_ENTRY (entry);

		/*
		 * check magic header
		 */

		if (entry->magic != BFS_INDEX_MAGIC) {

			/*
			 * bad magic header
			 */

			printk (KERN_ERR "BFS: "
				"magic header of index entry is bad value.\n");
#ifdef CONFIG_BFS_CONV
			putname (entry);
#endif
			brelse(bh);

			return NULL;
		}

		/*
		 * set index entry postion
		 */

		*offset = 0;
		iaddr = inode;
		if (entry->node_size < sb->u.bfs_sb.block_size) {

			/*
			 * exist index node into index entry's block
			 */

			*offset = entry->node_size;

		} else {

			iaddr.start += entry->node_size / sb->u.bfs_sb.block_size;
			iaddr.len -= entry->node_size / sb->u.bfs_sb.block_size;
			*offset = entry->node_size % sb->u.bfs_sb.block_size;

		}

#ifdef CONFIG_BFS_CONV
		putname (entry);
#endif
		brelse(bh);

	} else {
		BFS_OUTPUT ((" skip index node\n"));
		iaddr = inode;
	}

	BFS_OUTPUT ((" inode of index node\n"));
	BFS_DUMP_INODE_ADDR (iaddr);

	bh = bfs_bread2 (sb, iaddr);
	if (!bh) {
		return NULL;
	}

	BFS_OUTPUT (("<--- bfs_read_index_node() offset = %016x\n", *offset));

	return bh;
}


/*
 * bfs_get_key_from_index_node
 *
 * description:
 *  Get key form index node
 *
 *  parameter:
 *   bn     ... index node structure
 *   pos    ... key position
 *   fstype ... filesystem type
 *   name   ... keyname return
 *   len    ... length of keyname return.
 *   iaddr  ... key valude return.
 */

char * bfs_get_key_from_index_node (bfs_index_node * node, int *pos, int fstype,
	char * name, int * len, bfs_off_t * iaddr)
{
	char *           key;
	__u16 *          key_array;
	bfs_off_t *      key_value;
	int              key_offset;
	char *           namep = NULL;
	bfs_index_node * bn;

	BFS_OUTPUT (("---> bfs_get_key_from_index_node() "
		"pos %d\n", *pos));

	/*
	 * Convert little-edian and big-edian.
	 */
#ifndef CONFIG_BFS_CONV
	bn = node;
#else
	bn = (bfs_index_node *) __getname();
	bfs_convert_index_node (fstype, node, bn);
#endif

	if (*pos < 0 || bn->all_key_count <= *pos) {
		*pos = -1;
		return NULL;
	}

	/*
	 * Only key_array, byte is 8byte...
	 */

	key = (char*) node + sizeof(bfs_index_node);
	key_offset = (bn->all_key_length + sizeof(bfs_index_node)) % 8;
	key_array = (__u16 *) (key + bn->all_key_length
		+ (!key_offset ? 0 : 8 - key_offset));
	key_value = (bfs_inode_addr *) (((char *) key_array) + sizeof(__u16)
		* bn->all_key_count);
#ifdef BFS_DEBUG
	{
		int j;
		char key_test[4096];

		memcpy (key_test, key, bn->all_key_length);
		key_test[bn->all_key_length] = '\0';

		BFS_OUTPUT ((" key %08x key_array %08x key_value %08x\n",
			(long) key, (long) key_array, (long) key_value));

		BFS_OUTPUT ((" key content %s\n", key_test));

		for (j = 0; j < *pos; j++) {
			BFS_OUTPUT ((" kay_array[%d] = %Ld  ",
				j, key_array[j]));
			BFS_OUTPUT (("kay_value[%d] = %Ld\n", j, key_value[j]));
		}
	}
#endif
	switch (fstype)
	{
	case BFS_X86:
		if (*pos == 0) {
			namep = key;
			*len = le16_to_cpu(key_array[*pos]);
		} else {
			namep = key + le16_to_cpu(key_array[*pos - 1]);
			*len = le16_to_cpu(key_array[*pos])
				- le16_to_cpu(key_array[*pos - 1]);
		}
		*iaddr = le64_to_cpu(key_value[*pos]);
		break;

	case BFS_PPC:
		if (*pos == 0) {
			namep = key;
			*len = be16_to_cpu(key_array[*pos]);
		} else {
			namep = key + be16_to_cpu(key_array[*pos - 1]);
			*len = be16_to_cpu(key_array[*pos])
				- be16_to_cpu(key_array[*pos - 1]);
		}
		*iaddr = be64_to_cpu(key_value[*pos]);
		break;
	}

	/*
	 * copy from namep to name
	 */

	memcpy (name, namep, *len);
	name[*len] = '\0';
	(*pos)++;

#ifdef CONFIG_BFS_CONV
	putname (bn);
#endif

	BFS_OUTPUT (("<--- bfs_get_key_from_index_node() "
		"result key [%s], value %Lu\n", name, *iaddr));

	return name;
}
