/*
 *  linux/fs/bfs/namei.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 *
 *  from
 *
 *  linux/fs/minix/namei.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/quotaops.h>


/*
 * bfs_find_entry
 *
 * description:
 *  find file or directry entry
 *
 * parameter:
 *  dir     ... parent's inode
 *  name    ... file or directry name
 *  namelen ... length of name
 */

static bfs_off_t bfs_find_entry (struct inode * dir, const char * const name,
	int namelen)
{
	struct buffer_head *    bh;
	bfs_index_node *        bn;
	struct super_block *    sb = dir->i_sb;
	int                     sd_pos = 0;
	int                     key_pos;
	char                    key[BFS_NAME_LEN + 1];
	int                     key_len;
	bfs_off_t               value;

	BFS_OUTPUT (("---> bfs_find_entry()\n"));

	while (sd_pos >= 0) {
		bfs_data_stream * ds = &dir->u.bfs_i.i_data.ds;
		bfs_inode_addr    iaddr;
		int               flags;
		bfs_off_t         offset;

		/*
		 * get index node from small_data.
		 *
		 * First data stream has index entry, but other than it,
		 * doesn't have it.  
		 */

		if (sd_pos)
			flags = 1;
		else
			flags = 0;

		iaddr = bfs_read_data_stream (sb, ds, &sd_pos);
		if (sd_pos <= 0) {

			/*
			 * cannot read next data stream
			 */

			BFS_OUTPUT (("<--- bfs_find_entry() "
				"cannot read next data stream\n"));
			return 0;
		}

		bh = bfs_read_index_node (iaddr, sb, flags, &offset);
		if (!bh) {

			/*
			 * cannot read index node
			 */

			BFS_OUTPUT (("<--- bfs_find_entry() "
				"cannot read index node\n"));
			return 0;

		}

		bn = (bfs_index_node *) (bh->b_data + offset);

		/*
		 * explore in this block
		 */

		key_pos = 0;
		while (key_pos >= 0) {

			bfs_get_key_from_index_node (bn, &key_pos,
				sb->u.bfs_sb.mount_opts.bfs_type, key,
				&key_len, &value);
			if (key_pos < 0) {

				BFS_OUTPUT ((" not exist in this block\n"));
				break;

			}

			if (!strcmp (key, name)) {

				brelse (bh);

			BFS_OUTPUT (("<--- bfs_find_entry() "
				"value = %Lu\n", value));
			return value;

			}
		}

		brelse (bh);
	}

	BFS_OUTPUT (("<--- bfs_find_entry() not found\n"));
	return 0;
}



int bfs_lookup (struct inode * dir, struct dentry *dentry)
{
	struct inode * inode = NULL;
	bfs_off_t      offset;
	int            len;
	char *         tmpname;

	BFS_OUTPUT (("---> bfs_lookup() "
		"name %s inode %ld\n", dentry->d_name.name, dir->i_ino));

	/*
	 * Convert to UTF-8
	 */

	if (!bfs_nls2utf (dentry->d_name.name, dentry->d_name.len, &tmpname,
		&len, dir->i_sb)) {

		return -ENOMEM;
	}

	if (len > BFS_NAME_LEN) {
		putname (tmpname);
		return -ENAMETOOLONG;
	}

	offset = bfs_find_entry (dir, tmpname, len);
	putname (tmpname);

	if (offset) {

		inode = iget (dir->i_sb, (ino_t) offset);
		if (!inode) {

			return -EACCES;

		}
	}

	d_add (dentry, inode);

	BFS_OUTPUT (("<--- bfs_lookup()\n"));

	return 0;
}
